# Copyright (c) 2025 Oracle and/or its affiliates
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Implements the crash "ps" command for drgn.

This command displays process status for selected, or all, processes
in the system.
"""

import argparse
import collections
import functools
import sys
from typing import AbstractSet, Any, List, Optional, Sequence, Tuple

from drgn import FaultError, Object, Program
from drgn.commands import (
    CommandArgumentError,
    argument,
    drgn_argument,
    mutually_exclusive_group,
    parse_shell_command,
    unquote_shell_word,
)
from drgn.commands._crash.common import (
    _CRASH_FOREACH_SUBCOMMANDS,
    Cpuspec,
    CrashDrgnCodeBuilder,
    _crash_foreach_subcommand,
    _format_seconds_duration,
    _pid_or_task_or_command,
    _print_task_header,
    _TaskSelector,
    crash_custom_command,
    parse_cpuspec,
    print_task_header,
)
from drgn.helpers.common.format import CellFormat, escape_ascii_string, print_table
from drgn.helpers.linux.cpumask import for_each_online_cpu
from drgn.helpers.linux.kthread import task_is_kthread
from drgn.helpers.linux.list import list_for_each_entry
from drgn.helpers.linux.mm import (
    mm_cmdline,
    mm_environ,
    task_rss,
    task_vsize,
    totalram_pages,
)
from drgn.helpers.linux.pid import for_each_task_in_group
from drgn.helpers.linux.resource import task_rlimits
from drgn.helpers.linux.sched import (
    _TASK_STATE_CHAR_TO_STATE,
    cpu_rq,
    task_cpu,
    task_on_cpu,
    task_state_to_char,
)
from drgn.helpers.linux.timekeeping import ktime_get_coarse_ns

_SCHED_POLICIES = {
    "NORMAL": 0,
    # Crash doesn't accept this, but it's the standard name.
    "OTHER": 0,
    "FIFO": 1,
    "RR": 2,
    "BATCH": 3,
    # SCHED_ISO never made it upstream, but crash accepts it.
    "ISO": 4,
    "IDLE": 5,
    "DEADLINE": 6,
    "EXT": 7,
}


def _parse_sched_policies(arg: str) -> AbstractSet[int]:
    policies = set()
    for a in arg.split(","):
        try:
            policies.add(int(a, 10))
            continue
        except ValueError:
            pass

        try:
            policies.add(int(a, 16))
            continue
        except ValueError:
            pass

        try:
            policies.add(_SCHED_POLICIES[a.upper()])
        except KeyError:
            raise ValueError(f"invalid scheduling policy: {a}") from None
    return policies


def _format_nanosecond_duration(nanoseconds: int) -> str:
    days, nanoseconds = divmod(nanoseconds, 86400_000_000_000)
    hours, nanoseconds = divmod(nanoseconds, 3600_000_000_000)
    minutes, nanoseconds = divmod(nanoseconds, 60_000_000_000)
    seconds, nanoseconds = divmod(nanoseconds, 1_000_000_000)
    milliseconds = nanoseconds // 1_000_000
    return f"{days} {hours:02d}:{minutes:02d}:{seconds:02d}.{milliseconds:03d}"


def _ps_parents(task_selector: _TaskSelector, drgn_arg: bool) -> None:
    prog = task_selector.prog

    if drgn_arg:
        code = CrashDrgnCodeBuilder(prog)
        with task_selector.begin_task_loop(code):
            with code.begin_block("while True:\n"):
                code.append_task_header()
                code.append(
                    """\

parent = task.parent.read_()
if parent == task:
    break
task = parent
"""
                )
        return code.print()

    first = True
    for task in task_selector.tasks():
        if first:
            first = False
        else:
            print()

        parents = [task]
        while True:
            parent = task.parent.read_()
            if parent == task:
                break
            parents.append(parent)
            task = parent

        for level, task in enumerate(reversed(parents)):
            sys.stdout.write(" " * level)
            print_task_header(task)


def _ps_children(task_selector: _TaskSelector, drgn_arg: bool) -> None:
    prog = task_selector.prog

    if drgn_arg:
        code = CrashDrgnCodeBuilder(prog)
        with task_selector.begin_task_loop(code):
            code.append_task_header()
            code.add_from_import("drgn.helpers.linux.list", "list_for_each_entry")
            with code.begin_block(
                """\
for child in list_for_each_entry(
    "struct task_struct", task.children.address_of_(), "sibling"
):
"""
            ):
                code.append_task_header(variable="child")
        return code.print()

    first = True
    for task in task_selector.tasks():
        if first:
            first = False
        else:
            print()

        print_task_header(task)
        found = False
        for child in list_for_each_entry(
            "struct task_struct", task.children.address_of_(), "sibling"
        ):
            found = True
            sys.stdout.write("  ")
            print_task_header(child)
        if not found:
            print("  (no children)")


def _ps_times(task_selector: _TaskSelector, drgn_arg: bool) -> None:
    prog = task_selector.prog

    if drgn_arg:
        code = CrashDrgnCodeBuilder(prog)
        code.add_from_import("drgn.helpers.linux.timekeeping", "ktime_get_coarse_ns")
        code.append("now = ktime_get_coarse_ns()\n\n")
        with task_selector.begin_task_loop(code):
            code.append_task_header()
            code.append(
                """\
start_time = task.start_time
run_time = now - start_time
utime = task.utime
stime = task.stime
"""
            )
        return code.print()

    now = ktime_get_coarse_ns(prog)

    first = True
    for task in task_selector.tasks():
        if first:
            first = False
        else:
            print()

        print_task_header(task)
        start_time = task.start_time.value_()
        sys.stdout.write(
            f"""\
    RUN TIME: {_format_seconds_duration((now.value_() - start_time) // 1_000_000_000)}
  START TIME: {start_time}
       UTIME: {task.utime.value_()}
       STIME: {task.stime.value_()}
"""
        )


def _ps_last_arrival(
    task_selector: _TaskSelector,
    drgn_arg: bool,
    elapsed: bool,
    cpuspec: Optional[Cpuspec],
) -> None:
    prog = task_selector.prog

    if drgn_arg:
        code = CrashDrgnCodeBuilder(prog)
        with task_selector.begin_task_loop(code):
            if elapsed:
                code.add_from_import(
                    "drgn.helpers.linux.sched", "task_since_last_arrival_ns"
                )
                code.append("elapsed = task_since_last_arrival_ns(task)\n")
            else:
                code.append("last_arrival = task.sched_info.last_arrival\n")
            code.add_from_import("drgn.helpers.linux.sched", "task_state_to_char")
            code.append("state = task_state_to_char(task)\n")
            # append_task_header() already gets the CPU, so we don't bother
            # doing anything with the cpuspec.
            code.append_task_header()
        return code.print()

    rows = [
        (
            task.sched_info.last_arrival.value_(),
            task_cpu(task),
            task_state_to_char(task),
            task,
        )
        for task in task_selector.tasks()
    ]
    if elapsed:
        rq_clocks = {
            cpu: cpu_rq(prog, cpu).clock.value_() for cpu in for_each_online_cpu(prog)
        }
    rows.sort(reverse=True)

    def print_rows(rows: Sequence[Tuple[int, int, str, Object]]) -> None:
        timestamp_width = None

        for last_arrival, cpu, state, task in rows:
            if elapsed:
                timestamp = _format_nanosecond_duration(
                    max(rq_clocks[cpu] - last_arrival, 0)
                )
            else:
                timestamp = str(last_arrival)

            if timestamp_width is None:
                timestamp_width = len(timestamp)
            sys.stdout.write(f"[{timestamp:>{timestamp_width}}] [{state}]  ")
            _print_task_header(task, cpu=cpu)

    if cpuspec is None:
        print_rows(rows)
    else:
        first = True
        for cpu in cpuspec.cpus(prog):
            if first:
                first = False
            else:
                print()
            print(f"CPU: {cpu}")
            print_rows([row for row in rows if row[1] == cpu])


def _ps_arguments(task_selector: _TaskSelector, drgn_arg: bool) -> None:
    prog = task_selector.prog

    if drgn_arg:
        code = CrashDrgnCodeBuilder(prog)
        with task_selector.begin_task_loop(code):
            code.append_task_header()
            code.add_from_import("drgn.helpers.linux.mm", "cmdline", "environ")
            code.append(
                """\
arg = cmdline(task)
env = environ(task)
"""
            )
        return code.print()

    first = True
    for task in task_selector.tasks():
        if first:
            first = False
        else:
            print()

        print_task_header(task)

        mm = task.mm.read_()
        if not mm:
            print("  (no mm)")
            continue

        try:
            argv = mm_cmdline(mm)
        except FaultError as e:
            print("ps:", e)
        else:
            print(f"ARG: {' '.join([escape_ascii_string(arg) for arg in argv])}")

        try:
            envp = mm_environ(mm)
        except FaultError as e:
            print("ps:", e)
        else:
            if envp:
                prefix = "ENV:"
                for env in envp:
                    print(prefix, escape_ascii_string(env))
                    prefix = "    "
            else:
                print("ENV:")


def _ps_thread_groups(task_selector: _TaskSelector, drgn_arg: bool) -> None:
    prog = task_selector.prog

    if drgn_arg:
        code = CrashDrgnCodeBuilder(prog)
        with task_selector.begin_task_loop(code):
            code.append_task_header()
            code.add_from_import("drgn.helpers.linux.pid", "for_each_task_in_group")
            with code.begin_block("for thread in for_each_task_in_group(task):\n"):
                code.append_task_header(variable="thread")
        return code.print()

    first = True
    for task in task_selector.tasks():
        if first:
            first = False
        else:
            print()

        print_task_header(task)
        found = False
        for task in for_each_task_in_group(task):
            found = True
            sys.stdout.write("  ")
            print_task_header(task)
        if not found:
            print("  (no threads)")


def _ps_rlimit(task_selector: _TaskSelector, drgn_arg: bool) -> None:
    prog = task_selector.prog

    if drgn_arg:
        code = CrashDrgnCodeBuilder(prog)
        with task_selector.begin_task_loop(code):
            code.append_task_header()
            code.add_from_import("drgn.helpers.linux.resource", "task_rlimits")
            code.append("rlimits = task_rlimits(task)\n")
        return code.print()

    first = True
    for task in task_selector.tasks():
        if first:
            first = False
        else:
            print()

        print_task_header(task)
        print_table(
            [
                (
                    "",
                    CellFormat("RLIMIT", ">"),
                    CellFormat("CURRENT", "^"),
                    CellFormat("MAXIMUM", "^"),
                ),
                *(
                    (
                        "",
                        CellFormat(name, ">"),
                        CellFormat(
                            "(unlimited)" if limit.cur is None else limit.cur, "^"
                        ),
                        CellFormat(
                            "(unlimited)" if limit.max is None else limit.max, "^"
                        ),
                    )
                    for name, limit in task_rlimits(task).items()
                ),
            ]
        )


def _ps_summary(task_selector: _TaskSelector, drgn_arg: bool) -> None:
    prog = task_selector.prog

    if drgn_arg:
        code = CrashDrgnCodeBuilder(prog)
        code.add_import("collections")
        code.append("counter = collections.Counter()\n")
        with task_selector.begin_task_loop(code):
            code.add_from_import("drgn.helpers.linux.sched", "task_state_to_char")
            code.append("counter[task_state_to_char(task)] += 1\n")
        code.append(
            """\

for state, num in counter.items():
    ...
"""
        )
        return code.print()

    for state, num in collections.Counter(
        task_state_to_char(task) for task in task_selector.tasks()
    ).items():
        print(f"  {state}: {num}")


@_crash_foreach_subcommand(
    arguments=(
        argument(
            "-G",
            dest="group_leader",
            action="store_true",
        ),
        argument(
            "-y",
            dest="policy",
        ),
        mutually_exclusive_group(
            argument(
                "-s",
                dest="stack_pointer",
                action="store_true",
            ),
            argument(
                "-p",
                dest="func",
                action="store_const",
                const=_ps_parents,
            ),
            argument(
                "-c",
                dest="func",
                action="store_const",
                const=_ps_children,
            ),
            argument(
                "-t",
                dest="func",
                action="store_const",
                const=_ps_times,
            ),
            argument(
                "-l",
                dest="last_arrival_timestamp",
                action="store_true",
            ),
            argument(
                "-m",
                dest="last_arrival_elapsed",
                action="store_true",
            ),
            argument(
                "-a",
                dest="func",
                action="store_const",
                const=_ps_arguments,
            ),
            argument(
                "-g",
                dest="func",
                action="store_const",
                const=_ps_thread_groups,
            ),
            argument(
                "-r",
                dest="func",
                action="store_const",
                const=_ps_rlimit,
            ),
            argument(
                "-S",
                dest="func",
                action="store_const",
                const=_ps_summary,
            ),
        ),
        argument(
            "-C",
            dest="cpu",
        ),
        argument(
            "-H",
            dest="header",
            action="store_false",
        ),
        # Note: crash doesn't support foreach ps -S, -H, or -C, but supporting
        # them actually makes things easier for us, so we do.
        drgn_argument,
    ),
)
def _crash_foreach_ps(task_selector: _TaskSelector, args: argparse.Namespace) -> None:
    if args.last_arrival_timestamp or args.last_arrival_elapsed:
        cpuspec = None if args.cpu is None else parse_cpuspec(args.cpu)
    elif args.cpu is not None:
        raise CommandArgumentError("-C can only be used with -l or -m")

    if args.group_leader or args.func == _ps_thread_groups:
        task_selector._group_leader = True
    if args.policy is not None:
        task_selector._policies = _parse_sched_policies(args.policy)

    if args.func is not None:
        return args.func(task_selector, args.drgn)

    if args.last_arrival_timestamp or args.last_arrival_elapsed:
        return _ps_last_arrival(
            task_selector, args.drgn, args.last_arrival_elapsed, cpuspec
        )

    prog = task_selector.prog

    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)
        code.add_from_import(
            "drgn.helpers.linux.mm", "task_rss", "task_vsize", "totalram_pages"
        )
        code.append("total_mem = totalram_pages()\n\n")
        with task_selector.begin_task_loop(code):
            code.add_from_import(
                "drgn.helpers.linux.sched", "task_cpu", "task_state_to_char"
            )
            code.append(
                """\
pid = task.pid
ppid = task.parent.pid
cpu = task_cpu(task)
state = task_state_to_char(task)
rss = task_rss(task)
mem_usage = rss.total / total_mem
vsize = task_vsize(task)
comm = task.comm
"""
            )
        return code.print()

    rows: List[Sequence[Any]] = []
    if args.header:
        rows.append(
            (
                "",
                CellFormat("PID", ">"),
                CellFormat("PPID", ">"),
                CellFormat("CPU", ">"),
                CellFormat("KSTACKP" if args.stack_pointer else "TASK", "^"),
                "ST",
                CellFormat("%MEM", ">"),
                CellFormat("VSZ", ">"),
                CellFormat("RSS", ">"),
                "COMM",
            )
        )

    page_size = prog["PAGE_SIZE"].value_()
    total_mem = totalram_pages(prog)
    for task in task_selector.tasks():
        if args.stack_pointer:
            try:
                pointer_cell = CellFormat(prog.stack_trace(task)[0].sp, "^x")
            except ValueError:
                pointer_cell = CellFormat("--", "^")
        else:
            pointer_cell = CellFormat(task.value_(), "^x")

        rss = task_rss(prog, task)

        comm_column = escape_ascii_string(task.comm.string_(), escape_backslash=True)
        if task_is_kthread(task):
            comm_column = f"[{comm_column}]"

        rows.append(
            (
                ">" if task_selector._on_cpu or task_on_cpu(task) else "",
                task.pid.value_(),
                task.parent.pid.value_(),
                task_cpu(task),
                pointer_cell,
                task_state_to_char(task),
                CellFormat(rss.total / total_mem * 100, ".1f"),
                task_vsize(task) // 1024,
                rss.total * page_size // 1024,
                comm_column,
            )
        )
    print_table(rows)


@crash_custom_command(
    description="process information",
    long_description="display process status information",
    arguments=(
        mutually_exclusive_group(
            argument(
                "-k",
                dest="kernel",
                action="store_true",
                help="only display kernel threads",
            ),
            argument(
                "-u",
                dest="user",
                action="store_true",
                help="only display user tasks",
            ),
        ),
        argument(
            "-G",
            dest="group_leader",
            action="store_true",
            help="only display the thread group leader in each thread group",
        ),
        argument(
            "-y",
            dest="policy",
            help="""
            only display tasks with the given scheduling policy, as a
            comma-separated list of the following (case-insensitive) policy
            names or their integer values:

            |NORMAL or OTHER (0)
            |FIFO (1)
            |RR (2)
            |BATCH (3)
            |IDLE (5)
            |DEADLINE (6)
            |EXT (7)
            """,
        ),
        # In crash, -A is mutually exclusive with -p, -c, etc., but we have no
        # reason to make that restriction.
        argument(
            "-A",
            dest="active",
            action="store_true",
            help="display only the active task on each CPU",
        ),
        mutually_exclusive_group(
            argument(
                "-s",
                dest="stack_pointer",
                action="store_true",
                help="display the kernel stack pointer instead of the task_struct",
            ),
            argument(
                "-p",
                dest="func",
                action="store_const",
                const=_ps_parents,
                help="display the parental hierarchy of selected tasks",
            ),
            argument(
                "-c",
                dest="func",
                action="store_const",
                const=_ps_children,
                help="display the children of selected tasks",
            ),
            argument(
                "-t",
                dest="func",
                action="store_const",
                const=_ps_times,
                help="""
                display the task run time, start time, cumulative user time,
                and cumulative system time
                """,
            ),
            argument(
                "-l",
                dest="last_arrival_timestamp",
                action="store_true",
                help="""
                display the last_arrival timestamp of selected tasks, and sort
                from most-recent to least-recent
                """,
            ),
            argument(
                "-m",
                dest="last_arrival_elapsed",
                action="store_true",
                help="""
                like -l, but display the difference between the current
                runqueue clock and the last_arrival timestamp as
                "days hours:minutes:seconds.milliseconds"
                """,
            ),
            argument(
                "-a",
                dest="func",
                action="store_const",
                const=_ps_arguments,
                help="""
                display the command line arguments and environment variables of
                selected userspace tasks
                """,
            ),
            argument(
                "-g",
                dest="func",
                action="store_const",
                const=_ps_thread_groups,
                help="""
                display threads in the thread groups of selected tasks and
                group threads by thread group
                """,
            ),
            argument(
                "-r",
                dest="func",
                action="store_const",
                const=_ps_rlimit,
                help="display resource limits (rlimits) of selected tasks",
            ),
            argument(
                "-S",
                dest="func",
                action="store_const",
                const=_ps_summary,
                help="""
                display a summary of the number of selected tasks in each task
                state
                """,
            ),
        ),
        argument(
            "-C",
            dest="cpu",
            help="""
            for -l or -m, only display tasks on the given CPUs, which may be a
            comma-separated string of CPU numbers or ranges (e.g., '0,3-4'),
            and group selected tasks by CPU
            """,
        ),
        argument(
            "-H",
            dest="header",
            action="store_false",
            help="do not print a header line",
        ),
        argument(
            "tasks",
            metavar="pid|task|command",
            type=_pid_or_task_or_command,
            nargs="*",
            help=r"""
            display only this task, given as a decimal process ID, hexadecimal
            ``task_struct``, single-quoted (``'``) regular expression matching
            command names, or a literal command name (optionally prefixed with
            ``\`` to disambiguate it). May be given multiple times
            """,
        ),
        drgn_argument,
    ),
    parse=functools.partial(parse_shell_command, unquote=False),
)
def _crash_cmd_ps(
    prog: Program,
    name: str,
    quoted_args: Sequence[str],
    *,
    parser: argparse.ArgumentParser,
    **kwargs: Any,
) -> None:
    args = parser.parse_args(
        # Arguments starting with "'" or "\" have special meaning, so we don't
        # unquote those.
        [
            (
                arg
                if arg.startswith("'") or arg.startswith("\\")
                else unquote_shell_word(arg)
            )
            for arg in quoted_args
        ]
    )
    task_selector = _TaskSelector(
        prog,
        args.tasks,
        kernel=args.kernel,
        user=args.user,
        # Note: group_leader and policies are overridden in
        # _crash_foreach_ps().
        on_cpu=args.active,
        sort=True,
    )
    _crash_foreach_ps(task_selector, args)


@crash_custom_command(
    description="run command on multiple tasks",
    long_description="""
    Run the given command on all tasks matching the given constraints (or all
    tasks in the system if no constraints are given).
    """,
    usage=r"**foreach** [**\-\-drgn**] [*pid* | *task* | *name* ...] "
    "[**kernel** | **user** | **gleader**] [**active**] [*state*] "
    "*command* [*-option* ...]",
    arguments=(
        argument(
            "pid",
            help="run the command on the task with this decimal process ID",
        ),
        argument(
            "task",
            help="""
            run the command on the task with this hexadecimal task_struct
            address
            """,
        ),
        argument(
            "name",
            help=r"""
            run the command on tasks with the given name. May be prefixed with
            ``\`` to disambiguate it as a literal command name. If
            single-quoted (``'``), then it is treated as a regular expression
            """,
        ),
        argument(
            "kernel",
            help="run the command on kernel threads",
        ),
        argument(
            "user",
            help="run the command on user tasks",
        ),
        argument(
            "gleader",
            help="run the command on thread group leaders",
        ),
        argument(
            "active",
            help="run the command on the active task on each CPU",
        ),
        argument(
            "state",
            help='run the command on tasks in this state ("R", "D", etc.)',
        ),
        argument(
            "command",
            help="""
            run this command on the selected tasks. Currently, **bt**,
            **files**, **ps**, **set**, **sig**, **task**, **vm**, and **vtop**
            are supported
            """,
        ),
        argument(
            "-option",
            action="store_true",
            help="additional option to pass to the command",
        ),
        drgn_argument,
    ),
    parse=functools.partial(parse_shell_command, unquote=False),
)
def _crash_cmd_foreach(
    prog: Program,
    name: str,
    quoted_args: Sequence[str],
    *,
    parser: argparse.ArgumentParser,
    **kwargs: Any,
) -> Any:
    args = [
        (
            arg
            if arg.startswith("'") or arg.startswith("\\")
            else unquote_shell_word(arg)
        )
        for arg in quoted_args
    ]

    tasks = []
    kernel = False
    user = False
    group_leader = False
    on_cpu = False
    state = None
    command_args = []
    for i, arg in enumerate(args):
        if arg in _CRASH_FOREACH_SUBCOMMANDS:
            subcommand = _CRASH_FOREACH_SUBCOMMANDS[arg]
            command_args.extend(args[i + 1 :])
            break
        elif arg.startswith("-"):
            command_args.append(arg)
        elif arg == "kernel":
            if group_leader:
                parser.error("gleader and kernel are mutually exclusive")
            if user:
                parser.error("user and kernel are mutually exclusive")
            kernel = True
        elif arg == "user":
            if kernel:
                parser.error("kernel and user are mutually exclusive")
            user = True
        elif arg == "gleader":
            if kernel:
                parser.error("kernel and gleader are mutually exclusive")
            user = group_leader = True
        elif arg == "active":
            on_cpu = True
        elif arg in _TASK_STATE_CHAR_TO_STATE:
            if state is not None:
                parser.error("only one task state allowed")
            state = arg
        else:
            tasks.append(_pid_or_task_or_command(arg))
    else:
        parser.error("no command given")

    return subcommand.func(
        _TaskSelector(
            prog,
            tasks,
            kernel=kernel,
            user=user,
            group_leader=group_leader,
            on_cpu=on_cpu,
            state=state,
        ),
        subcommand.parser.parse_args(command_args),
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
CPU Masks
---------

The ``drgn.helpers.linux.cpumask`` module provides helpers for working with CPU
masks from :linux:`include/linux/cpumask.h`.
"""

import operator
from typing import Iterator

from drgn import IntegerLike, Object, Program, cast, sizeof
from drgn.helpers.common.prog import takes_program_or_default
from drgn.helpers.linux.bitmap import bitmap_weight
from drgn.helpers.linux.bitops import for_each_set_bit

__all__ = (
    "cpu_online_mask",
    "cpu_possible_mask",
    "cpu_present_mask",
    "cpumask_of",
    "cpumask_to_cpulist",
    "cpumask_weight",
    "for_each_cpu",
    "for_each_online_cpu",
    "for_each_possible_cpu",
    "for_each_present_cpu",
    "num_online_cpus",
    "num_possible_cpus",
    "num_present_cpus",
)


# Before Linux kernel commit c4c54dd1caf1 ("kernel/cpu.c: change type of
# cpu_possible_bits and friends") (in v4.5), the CPU masks are struct cpumask
# *cpu_foo_mask instead of struct cpumask __cpu_foo_mask.
@takes_program_or_default
def cpu_online_mask(prog: Program) -> Object:
    """
    Return the mask of online CPUs.

    :return: ``struct cpumask *``
    """
    try:
        return prog["__cpu_online_mask"].address_of_()
    except KeyError:
        return prog["cpu_online_mask"]


@takes_program_or_default
def cpu_possible_mask(prog: Program) -> Object:
    """
    Return the mask of possible CPUs.

    :return: ``struct cpumask *``
    """
    try:
        return prog["__cpu_possible_mask"].address_of_()
    except KeyError:
        return prog["cpu_possible_mask"]


@takes_program_or_default
def cpu_present_mask(prog: Program) -> Object:
    """
    Return the mask of present CPUs.

    :return: ``struct cpumask *``
    """
    try:
        return prog["__cpu_present_mask"].address_of_()
    except KeyError:
        return prog["cpu_present_mask"]


@takes_program_or_default
def cpumask_of(prog: Program, cpu: IntegerLike) -> Object:
    """
    Return a mask containing only the given CPU.

    :return: ``struct cpumask *``
    """
    BITS_PER_LONG = sizeof(prog.type("unsigned long")) * 8
    cpu = operator.index(cpu)
    return cast(
        "struct cpumask *",
        prog["cpu_bit_bitmap"][1 + cpu % BITS_PER_LONG] - cpu // BITS_PER_LONG,
    )


def for_each_cpu(mask: Object) -> Iterator[int]:
    """
    Iterate over all of the CPUs in the given mask.

    :param mask: ``struct cpumask *``
    """
    try:
        nr_cpu_ids = mask.prog_["nr_cpu_ids"].value_()
    except KeyError:
        nr_cpu_ids = 1
    return for_each_set_bit(mask.bits, nr_cpu_ids)


@takes_program_or_default
def for_each_online_cpu(prog: Program) -> Iterator[int]:
    """Iterate over all online CPUs."""
    return for_each_cpu(cpu_online_mask(prog))


@takes_program_or_default
def for_each_possible_cpu(prog: Program) -> Iterator[int]:
    """Iterate over all possible CPUs."""
    return for_each_cpu(cpu_possible_mask(prog))


@takes_program_or_default
def for_each_present_cpu(prog: Program) -> Iterator[int]:
    """Iterate over all present CPUs."""
    return for_each_cpu(cpu_present_mask(prog))


def cpumask_weight(mask: Object) -> int:
    """
    Return the number of CPUs in the given mask

    :param mask: ``struct cpumask *``
    """
    try:
        nr_cpu_ids = mask.prog_["nr_cpu_ids"].value_()
    except KeyError:
        nr_cpu_ids = 1
    return bitmap_weight(mask.bits, nr_cpu_ids)


@takes_program_or_default
def num_online_cpus(prog: Program) -> int:
    """Return the number of online CPUs."""
    # Since Linux kernel commit 0c09ab96fc82 ("cpu/hotplug: Cache number of
    # online CPUs") (in v5.4), there's a cached atomic_t __num_online_cpus, but
    # it's not worth it to do the version detection.
    return cpumask_weight(cpu_online_mask(prog))


@takes_program_or_default
def num_possible_cpus(prog: Program) -> int:
    """Return the number of possible CPUs."""
    return cpumask_weight(cpu_possible_mask(prog))


@takes_program_or_default
def num_present_cpus(prog: Program) -> int:
    """Return the number of present CPUs."""
    return cpumask_weight(cpu_present_mask(prog))


def cpumask_to_cpulist(mask: Object) -> str:
    """
    Return a CPU mask as a CPU list string.

    >>> cpumask_to_cpulist(mask)
    0-3,8-11

    :param mask: ``struct cpumask *``
    :return: String in the `CPU list format
        <https://man7.org/linux/man-pages/man7/cpuset.7.html#FORMATS>`_.
    """
    start = end = -2
    parts = []
    for cpu in for_each_cpu(mask):
        if cpu == end + 1:
            end = cpu
        else:
            if start >= 0:
                parts.append(str(start) if start == end else f"{start}-{end}")
            start = end = cpu
    if start >= 0:
        parts.append(str(start) if start == end else f"{start}-{end}")
    return ",".join(parts)



#!/usr/bin/env drgn
# Copyright (c) SUSE Linux.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""A simplified implementation of mount(1) using drgn"""

from drgn.helpers.linux.fs import for_each_mount, mount_dst, mount_fstype, mount_src

print("Mount            Type         Devname      Dirname")
for mount in for_each_mount(prog):
    maddr = mount.value_()
    src = mount_src(mount).decode()
    dst = mount_dst(mount).decode()
    type_ = mount_fstype(mount).decode()

    print(f"{maddr:<16x} {type_:<12} {src:<12} {dst}")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Internal utilities for drgn

This package contains utilities shared between the drgn package and supporting
build/test code. You should not use them.

This package must not depend on the drgn package itself since it is used before
the _drgn extension module is built.
"""



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
BPF
---

The ``drgn.helpers.linux.bpf`` module provides helpers for working with BPF
interface in :linux:`include/linux/bpf.h`, :linux:`include/linux/bpf-cgroup.h`,
etc.
"""


import itertools
from typing import Iterator

from drgn import IntegerLike, Object, Program, cast
from drgn.helpers.common.prog import takes_program_or_default
from drgn.helpers.linux.idr import idr_find, idr_for_each
from drgn.helpers.linux.list import hlist_for_each_entry, list_for_each_entry

__all__ = (
    "bpf_btf_for_each",
    "bpf_link_for_each",
    "bpf_map_for_each",
    "bpf_prog_for_each",
    "cgroup_bpf_prog_for_each",
    "cgroup_bpf_prog_for_each_effective",
    "bpf_prog_used_maps",
    "bpf_prog_by_id",
    "bpf_map_by_id",
)


@takes_program_or_default
def bpf_btf_for_each(prog: Program) -> Iterator[Object]:
    """
    Iterate over all BTF objects.

    This is only supported since Linux v4.18.

    :return: Iterator of ``struct btf *`` objects.
    """
    type = prog.type("struct btf *")
    # BTF was introduced in Linux kernel commit 69b693f0aefa ("bpf: btf:
    # Introduce BPF Type Format (BTF)") (in v4.18). btf_idr was added in a
    # later commit in v4.18, 78958fca7ead ("bpf: btf: Introduce BTF ID").
    for nr, entry in idr_for_each(prog["btf_idr"]):
        yield cast(type, entry)


@takes_program_or_default
def bpf_link_for_each(prog: Program) -> Iterator[Object]:
    """
    Iterate over all BPF links.

    This is only supported since Linux v5.8.

    :return: Iterator of ``struct bpf_link *`` objects.
    """
    type = prog.type("struct bpf_link *")
    # link_idr didn't exist before Linux kernel commit a3b80e107894 ("bpf:
    # Allocate ID for bpf_link") (in v5.8). struct bpf_link didn't exist at all
    # before Linux kernel commit 70ed506c3bbc ("bpf: Introduce pinnable
    # bpf_link abstraction") (in v5.7), and we don't support Linux v5.7
    # anyways.
    for nr, entry in idr_for_each(prog["link_idr"]):
        yield cast(type, entry)


@takes_program_or_default
def bpf_map_for_each(prog: Program) -> Iterator[Object]:
    """
    Iterate over all BPF maps.

    This is only supported since Linux v4.13.

    :return: Iterator of ``struct bpf_map *`` objects.
    """
    type = prog.type("struct bpf_map *")
    # map_idr didn't exist before Linux kernel commit f3f1c054c288 ("bpf:
    # Introduce bpf_map ID") (in v4.13).
    for nr, entry in idr_for_each(prog["map_idr"]):
        yield cast(type, entry)


@takes_program_or_default
def bpf_prog_for_each(prog: Program) -> Iterator[Object]:
    """
    Iterate over all BPF programs.

    This is only supported since Linux v4.13.

    :return: Iterator of ``struct bpf_prog *`` objects.
    """
    type = prog.type("struct bpf_prog *")
    # prog_idr didn't exist before Linux kernel commit dc4bb0e23561 ("bpf:
    # Introduce bpf_prog ID") (in v4.13).
    for nr, entry in idr_for_each(prog["prog_idr"]):
        yield cast(type, entry)


def cgroup_bpf_prog_for_each(
    cgrp: Object, bpf_attach_type: IntegerLike
) -> Iterator[Object]:
    """
    Iterate over all cgroup BPF programs of the given attach type attached to
    the given cgroup.

    :param cgrp: ``struct cgroup *``
    :param bpf_attach_type: ``enum cgroup_bpf_attach_type`` (``enum
        bpf_attach_type`` before Linux 5.15)
    :return: Iterator of ``struct bpf_prog *`` objects.
    """
    # Before Linux kernel commit 3007098494be ("cgroup: add support for eBPF
    # programs") (in v4.10), struct cgroup::bpf didn't exist because cgroup BPF
    # programs didn't exist.
    try:
        cgrp_bpf = cgrp.bpf
    except AttributeError:
        return
    # Since Linux kernel commit 324bda9e6c5a ("bpf: multi program support for
    # cgroup+bpf") (in v4.15), the attached programs are stored in an array of
    # lists, struct cgroup_bpf::progs. Before that, only one program of each
    # attach type could be attached to a cgroup, so the attached programs are
    # stored in an array of struct bpf_prog *, struct cgroup_bpf::prog.
    try:
        progs = cgrp_bpf.progs
    except AttributeError:
        # If the kernel was not configured with CONFIG_CGROUP_BPF, then struct
        # cgroup_bpf is an empty structure.
        try:
            prog = cgrp_bpf.prog[bpf_attach_type]
        except AttributeError:
            return
        if prog:
            yield prog
    else:
        # Since Linux kernel commit 00442143a2ab ("bpf: convert
        # cgroup_bpf.progs to hlist") (in v6.0-rc1), the list of programs is an
        # hlist_head. Before that, it was a list_head.
        list = progs[bpf_attach_type].address_of_()
        if hasattr(list, "first"):
            iterator = hlist_for_each_entry
        else:
            iterator = list_for_each_entry
        for pl in iterator("struct bpf_prog_list", list, "node"):
            yield pl.prog


def cgroup_bpf_prog_for_each_effective(
    cgrp: Object, bpf_attach_type: IntegerLike
) -> Iterator[Object]:
    """
    Iterate over all effective cgroup BPF programs of the given attach type for
    the given cgroup.

    :param cgrp: ``struct cgroup *``
    :param bpf_attach_type: ``enum bpf_attach_type``
    :return: Iterator of ``struct bpf_prog *`` objects.
    """
    # Before Linux kernel commit 3007098494be ("cgroup: add support for eBPF
    # programs") (in v4.10), struct cgroup::bpf didn't exist because cgroup BPF
    # programs didn't exist. Since then, if the kernel was not configured with
    # CONFIG_CGROUP_BPF, then struct cgroup_bpf is an empty structure.
    try:
        effective = cgrp.bpf.effective[bpf_attach_type]
    except AttributeError:
        return
    # Since Linux kernel commit 324bda9e6c5a ("bpf: multi program support for
    # cgroup+bpf") (in v4.15), struct cgroup_bpf::effective is an array of
    # struct bpf_prog_array. Before that, only one program of each attach type
    # could be effective for a cgroup, so struct cgroup_bpf::effective is an
    # array of struct bpf_prog *.
    try:
        effective_items = effective.items
    except AttributeError:
        if effective:
            yield effective
    else:
        for i in itertools.count():
            prog = effective_items[i].prog.read_()
            if not prog:
                break
            yield prog


def bpf_prog_used_maps(bpf_prog: Object) -> Iterator[Object]:
    """
    Yield maps used by a BPF program.

    :param bpf_prog: ``struct bpf_prog *``
    :return: Iterator of ``struct bpf_map *`` objects.
    """
    aux = bpf_prog.aux.read_()
    return iter(aux.used_maps[: aux.used_map_cnt])


@takes_program_or_default
def bpf_prog_by_id(prog: Program, id: IntegerLike) -> Object:
    """
    Get a BPF program by ID.

    This is only supported since Linux v4.13.

    :param prog: Program object
    :param id: BPF program ID
    :return: ``struct bpf_prog *`` object, or a null pointer if not found
    """
    return cast("struct bpf_prog *", idr_find(prog["prog_idr"].address_of_(), id))


@takes_program_or_default
def bpf_map_by_id(prog: Program, id: IntegerLike) -> Object:
    """
    Get a BPF map by ID.

    This is only supported since Linux v4.13.

    :param prog: Program object
    :param id: BPF map ID
    :return: ``struct bpf_map *`` object, or a null pointer if not found
    """
    return cast("struct bpf_map *", idr_find(prog["map_idr"].address_of_(), id))



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Maple Trees
-----------

The ``drgn.helpers.linux.mapletree`` module provides helpers for working with
maple trees from :linux:`include/linux/maple_tree.h`.

Maple trees were introduced in Linux 6.1.
"""

import collections
import operator
from typing import Iterator, Tuple

from drgn import NULL, IntegerLike, Object, Program, sizeof
from drgn.helpers.linux.xarray import _XA_ZERO_ENTRY, _xa_is_node, xa_is_zero

__all__ = (
    "mt_for_each",
    "mtree_load",
)


def _ulong_max(prog: Program) -> int:
    return (1 << (8 * sizeof(prog.type("unsigned long")))) - 1


# Combination of mte_to_node(), mte_node_type(), ma_data_end(), and
# ma_is_leaf().
def _mte_to_node(
    prog: Program, entry_value: int, max: int
) -> Tuple[Object, Object, Object, int, bool]:
    MAPLE_NODE_MASK = 255
    MAPLE_NODE_TYPE_MASK = 0xF
    MAPLE_NODE_TYPE_SHIFT = 0x3
    maple_leaf_64 = 1
    maple_range_64 = 2
    maple_arange_64 = 3

    node = Object(prog, "struct maple_node *", entry_value & ~MAPLE_NODE_MASK)
    type = (entry_value >> MAPLE_NODE_TYPE_SHIFT) & MAPLE_NODE_TYPE_MASK
    if type == maple_arange_64:
        m = node.ma64
        pivots = m.pivot
        slots = m.slot
        end = m.meta.end.value_()
    elif type == maple_range_64 or type == maple_leaf_64:
        m = node.mr64
        pivots = m.pivot
        slots = m.slot
        pivot = pivots[len(pivots) - 1].value_()
        if not pivot:
            end = m.meta.end.value_()
        elif pivot == max:
            end = len(pivots) - 1
        else:
            end = len(pivots)
    else:
        raise NotImplementedError(f"unknown maple_type {type}")

    return node, pivots, slots, end, type < maple_range_64


def mtree_load(mt: Object, index: IntegerLike, *, advanced: bool = False) -> Object:
    """
    Look up the entry at a given index in a maple tree.

    >>> entry = mtree_load(task.mm.mm_mt.address_of_(), 0x55d65cfaa000)
    >>> cast("struct vm_area_struct *", entry)
    *(struct vm_area_struct *)0xffff97ad82bfc930 = {
        ...
    }

    :param mt: ``struct maple_tree *``
    :param index: Entry index.
    :param advanced: Whether to return nodes only visible to the maple tree
        advanced API. If ``False``, zero entries (see
        :func:`~drgn.helpers.linux.xarray.xa_is_zero()`) will be returned as
        ``NULL``.
    :return: ``void *`` found entry, or ``NULL`` if not found.
    """
    prog = mt.prog_
    index = operator.index(index)
    entry = mt.ma_root.read_()
    entry_value = entry.value_()
    if _xa_is_node(entry_value):
        max = _ulong_max(prog)
        while True:
            node, pivots, slots, end, leaf = _mte_to_node(prog, entry_value, max)

            for offset in range(end):
                pivot = pivots[offset].value_()
                if pivot >= index:
                    max = pivot
                    break
            else:
                offset = end

            entry_value = slots[offset].value_()
            if leaf:
                if not advanced and entry_value == _XA_ZERO_ENTRY:
                    return NULL(prog, "void *")
                return Object(prog, "void *", entry_value)
    elif entry_value and index == 0:
        return entry
    else:
        return NULL(prog, "void *")


def mt_for_each(
    mt: Object, *, advanced: bool = False
) -> Iterator[Tuple[int, int, Object]]:
    """
    Iterate over all of the entries and their ranges in a maple tree.

    >>> for first_index, last_index, entry in mt_for_each(task.mm.mm_mt.address_of_()):
    ...     print(hex(first_index), hex(last_index), entry)
    ...
    0x55d65cfaa000 0x55d65cfaafff (void *)0xffff97ad82bfc930
    0x55d65cfab000 0x55d65cfabfff (void *)0xffff97ad82bfc0a8
    0x55d65cfac000 0x55d65cfacfff (void *)0xffff97ad82bfc000
    0x55d65cfad000 0x55d65cfadfff (void *)0xffff97ad82bfcb28
    ...

    :param mt: ``struct maple_tree *``
    :param advanced: Whether to return nodes only visible to the maple tree
        advanced API. If ``False``, zero entries (see
        :func:`~drgn.helpers.linux.xarray.xa_is_zero()`) will be skipped.
    :return: Iterator of (first_index, last_index, ``void *``) tuples. Both
        indices are inclusive.
    """
    entry = mt.ma_root.read_()
    entry_value = entry.value_()
    if _xa_is_node(entry_value):
        prog = mt.prog_
        queue = collections.deque(((entry_value, 0, _ulong_max(prog)),))
        while queue:
            entry_value, min, max = queue.popleft()
            node, pivots, slots, end, leaf = _mte_to_node(prog, entry_value, max)

            if leaf:
                prev = min
                for offset in range(end):
                    pivot = pivots[offset].value_()
                    slot = slots[offset].read_()
                    if slot and (advanced or not xa_is_zero(slot)):
                        yield (prev, pivot, slot)
                    prev = pivot + 1
                slot = slots[end].read_()
                if slot and (advanced or not xa_is_zero(slot)):
                    yield (prev, max, slot)
            else:
                prev = min
                for offset in range(end):
                    pivot = pivots[offset].value_()
                    queue.append((slots[offset].value_(), prev, pivot))
                    prev = pivot + 1
                queue.append((slots[end].value_(), prev, max))
    elif entry_value:
        yield (0, 0, entry)
